/*--------------------------------------------------------------------.
| menus.c							      |
| 								      |
| Copyright (c) 1997 Jonathan Hudson				      |
| 								      |
| You may distribute under the terms of either the GNU General Public |
| License or the Artistic License, as specified in the README file.   |
`--------------------------------------------------------------------*/

static char rcsid[] = "$Id$";

/*
 * $Log$ 
 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <qdos.h>
#include <qptr.h>
#include <unistd.h>
#include <signal.h>
#include "qfm.h"

extern void menu_button (WM_wwork_t *, char *);
int GetQM (WM_wwork_t *wwk, int len, char *p0, char *p1, char *p2);

typedef struct WM_menw WM_menw_t;

static long MEXEC (WM_wwork_t *, WM_menw_t *, char *, short, short);
static long MDRAW (WM_wwork_t *, WM_menw_t *);

static WM_action_t mdraw = {JSR, wm_drwaw, MDRAW};
static WM_action_t mexec = {JSR, wm_actme, MEXEC};

#define MAX_SECT    1
#define EXTRAS      1

static long MMOVE (struct WM_wwork *, struct WM_litm *li);
static struct WM_action mmove = {JSR, wm_actli, MMOVE};

static long MSLEEP (struct WM_wwork *, struct WM_litm *li);
static struct WM_action msleep = {JSR, wm_actli, MSLEEP};

/*-----------------------------.
| * Reset a selected menu item |
`-----------------------------*/

long lit_reset (WM_wwork_t * wwk, short lit)
{
    wwk->wstat->litem[lit] = WSI_MKAV;	/* set make available */
    return wm_ldraw (wwk, -1);	/* and re-draw */
}

static long MMOVE (struct WM_wwork *wwk, struct WM_litm *li)
{
    return PT__WMOV;
}

static long MSLEEP (struct WM_wwork *wwk, struct WM_litm *li)
{
    return PT__ZZZZ;
}

/*--------------------------------------------------------------------------.
|  Check if the menu fits in its window, and add scroll bars as necessary.  |
| Taken almost verbatim from the TT/c68 libqptr examples.		    |
`--------------------------------------------------------------------------*/

static void menu_chks (WM_menw_t * menu)
{
    WM_pwcb_t *pwcb;
    WM_splst_t *xl, *yl;

    short spacing;
    int i;

    pwcb = menu->ypwcb;		/* part window control block (scroll) */
    spacing = -menu->ys.c._spce;

    if (menu->nrow > menu->ysize / spacing)
    {
	if (pwcb->nsec == 0)
	    menu->xsize -= WW_XBAR;	/* adjust width for bar */
	pwcb->nsec = 1;		/* one section */
	pwcb->s[0].stt = 0;	/* from zero */
	pwcb->s[0].siz = ((menu->ysize - WW_YARROW * 2) / spacing);	/* size */
	menu->yoff = WW_YARROW;	/* offset for arrows */
    }
    else
    {
	if (pwcb->nsec)
	    menu->xsize += WW_XBAR;	/* adjust width for no bar */
	pwcb->nsec = 0;		/* no sections */
	menu->yoff = 0;		/* no arrows, no offset */
    }
    menu->xs.c._spce = -(menu->xsize) / menu->ncol;	/* width of each column */
    menu->xs.c._size = menu->xs.c._spce + menu->curw * 4;	/* ... item border */
}

/*---------------------------------------------------------.
|  Remove scroll bars from menu				   |
| Taken almost verbatim from the TT/c68 libqptr examples.  |
`---------------------------------------------------------*/

static void menu_rems (WM_menw_t * menu)
{
    WM_pwcb_t *pwcb;

    pwcb = menu->ypwcb;		/* part window control block (scroll) */

    if (pwcb->nsec)
    {
	menu->xsize += WW_XBAR;	/* adjust width for no bar */
	pwcb->nsec = 0;		/* no sections */
	menu->yoff = 0;
    }
}

static WM_wwork_t *ww = NULL;
static char *Mmstt = NULL;
extern int pick(int,int,int,int);
extern int ualarm(int);
extern int execute(WM_wwork_t*, void *, char);
extern short served(char *);

void signalTimer(int sig)
{
    static int n = -1;

    if(ww && Mmstt)
    {
	int m = served("QFAX");
	if(m != n)
	{
	    n = m;
	    *Mmstt = (m) ? WSI_AVBL : WSI_UNAV;
	    pick(ww->chid, 0, getpid(), K_WAKE);
	}
    }
    ualarm(50);
}

#if 0
static void showql (char *ctl,...)
{
    static int f = 0;
    va_list va;
    char t[80];


    if(f == 0) f = io_open("win2_tmp_ql.log", NEW_OVER);
    va_start (va, ctl);
    vsprintf (t, ctl, va);
    va_end (va);
    io_sstrg (f, -1, t, strlen (t));
}
#endif

/*-----------------------------------------------------------------------------.
| Load up a QPTR menu structure from our internal menu definition. Note that   |
| if the menu contains SEParators, then these built up here, as we didn't know |
| how wide they needed to be previously.				       |
`-----------------------------------------------------------------------------*/

static void LoadMenu (WM_menw_t * menu, MENUS * m)
{
    int i;
    WM_rowl_t *rowl;
    WM_mobj_t *mobj;
    char *mstt;
    int nitem;
    int nsize;
    ITEMS *im;

    if (menu->rowl)
    {
	xfree (menu->rowl);
    }

    menu->rowl = NULL;
    menu->ncol = 1;

    menu->nrow = nitem = m->height;

    nsize = (sizeof (WM_rowl_t) + sizeof (WM_mobj_t) +
	     sizeof (WM_mobj_t) + 1) * nitem;

    menu->rowl =		/* allocate all working space in a lump */
	rowl = (WM_rowl_t *) xalloc
	((int) ((sizeof (WM_rowl_t) + sizeof (WM_mobj_t) + 1) * nitem));
    mobj = (WM_mobj_t *) (((char *) rowl) + sizeof (WM_rowl_t) * nitem);
    menu->mstt =
	mstt = ((char *) mobj) + sizeof (WM_mobj_t) * nitem;

    im = m->items;
    for (i = 0; i < menu->nrow; i++)	/* initialise menu structures */
    {
	mobj->xjst = 1;
	mobj->yjst = 0;
	mobj->type = 0;
	mobj->skey = 0;
	mobj->item = i;

	if (im->l->act == SEP)
	{
	    *mstt = WSI_UNAV;
	    if (im->l->len == 0)
	    {
		short j;
		MINFO *lp = im->l;

		im->l = xalloc (sizeof (MINFO) + m->width + EXTRAS);
		im->l->act = SEP;
		im->l->len = m->width + EXTRAS;
		if (*lp->chrs)
		{
		    short ns = strlen (lp->chrs);
		    im->l->chrs[0] = '-';
		    strcpy (im->l->chrs + 1, lp->chrs);
		    j = ns + 1;
		}
		else
		{
		    j = 0;
		}
		for (; j < m->width + EXTRAS; j++)
		{
		    im->l->chrs[j] = '-';
		}
		xfree (lp);
	    }
	    mobj->pobj = &im->l->len;
	}
	else
	{
	    mobj->pobj = &im->l->len;
	    *mstt = WSI_AVBL;
	    mobj->pact = &mexec;
	    if (im->l->skey > 0)
	    {
		mobj->skey = im->l->skey;
	    }
	    else if (im->l->skey < 0)
	    {
		mobj->type = im->l->skey;
		mobj->skey = *(((struct QD_text *) mobj->pobj)->chrs
			       - (im->l->skey + 1));
		if (mobj->skey > 'Z')
		{
		    mobj->skey -= ' ';
		}
	    }
	    else
	    {
		*mstt = WSI_UNAV;
		Mmstt = mstt;
	    }
	}

	mstt++;
	rowl->rows = mobj++;	/* one object per row */
	(rowl++)->rowe = mobj;
	im = im->next;		/* next text */
    }
}


int menu (int chid, MENUS * m)
{

    short pr_w = 0;
    WM_wstat_t *ws;
    WM_info_t *info;
    WM_infw_t *infw;
    WM_menw_t *mw;
    char rbuf[2];

    /*--------------------------------------------------------------.
    | We only really need 2 of these, however, passing data back up |
    | from menu action routines is quite difficult, so this is a    |
    | convenient hack to hold a pointer.			    |
    | 								    |
    | I don't recommend you really write programs like this	    |
    `--------------------------------------------------------------*/

    WM_appl_t appl[3];
    WM_litm_t *ppl;
    int rval;
    WM_PWCB (2) pwcy;
    QD_text_t *ptitle;
    short xsize, ysize;
    short xo, yo;
    static short dirn;
    struct sigaction sac;

    /*--------------------------.
    | Set up the signal handler |
    `--------------------------*/

    sigemptyset(&(sac.sa_mask));
    sac.sa_handler=signalTimer;
    sigaction(SIGALRM,&sac,NULL);

    /* Check how big the menu can be  */

    /*----------------------------------------------------------------.
    | dynamically allocate space for all the data structures required |
    | for PTR menus. This is painfully necessary as we're called      |
    | recursively.						      |
    `----------------------------------------------------------------*/

    ppl = xalloc (3 * sizeof (WM_litm_t));
    info = xalloc (3 * sizeof (WM_info_t));
    infw = xalloc (2 * sizeof (WM_infw_t));
    ww = xalloc (sizeof (WM_wwork_t));
    mw = xalloc (sizeof (WM_menw_t));
    ws = xalloc (sizeof (WM_wstat_t));

    /*-------------------------------.
    | Now set up the menu structures |
    `-------------------------------*/

    (ppl+0)->xsize = 18;
    (ppl+0)->ysize = 10;
    (ppl+0)->xorg = 2;
    (ppl+0)->yorg = 2;
    (ppl+0)->type = 2;
    (ppl+0)->skey = K_MOVE;
    (ppl+0)->pobj = &mes_move;
    (ppl+0)->pact = &mmove;

    (ppl+1)->xsize = 24;
    (ppl+1)->ysize = 10;
    (ppl+1)->yorg = 2;
    (ppl+1)->type = 2;
    (ppl+1)->skey = K_SLEEP;
    (ppl+1)->pobj = &mes_slee;
    (ppl+1)->pact = &msleep;
    (ppl+1)->item = 1;

    (ppl + 2)->xsize = (ppl + 2)->ysize = -1;

    memset (&pwcy, '\0', sizeof (pwcy));

    (info + 0)->ysize = 10;
    (info + 0)->xorg = 2;
    (info + 0)->type = TYP_TEXT;
    (info + 0)->attr.t.ink = 1;

    (info + 1)->xsize = (info + 0)->ysize = -1;

    (infw + 0)->ysize = 14;
    (infw + 0)->papr = 84;

    (infw + 1)->ysize = 10;
    (infw + 1)->xorg = 24;
    (infw + 1)->yorg = 2;
    (infw + 1)->papr = 7;
    (infw + 1)->pobl = info;

    (infw + 2)->xsize = (infw + 2)->ysize = -1;

    mw->xorg = 2;
    mw->yorg = 14;
    mw->borw = 1;
    mw->borc = colarray[COLBORD];
    mw->papr = colarray[COLPAPR];
    mw->draw = &mdraw;
    mw->hit = wm__mhit;
    mw->ctrl = wm__pnsc;

    mw->ypsac = colarray[COLPSAC];
    mw->ypsbc = colarray[COLPSBC];
    mw->ypssc = colarray[COLPSCC];

    mw->curw = 1;
    mw->curc = colarray[COLCURC];
    mw->uback = colarray[COLUBACK];
    mw->uink = colarray[COLUINK];
    mw->aback = colarray[COLABACK];
    mw->aink = colarray[COLAINK];
    mw->sback = colarray[COLSBACK];
    mw->sink = colarray[COLSINK];
    mw->ncol = 1;
    mw->ys.c._size = -10;
    mw->ys.c._spce = -12;

    (appl + 0)->pappw = (WM_appw_t *) mw;
    (appl + 1)->pappw = NULL;

    ww->flag = NO_WCLEAR + 2;
    ww->borw = 1;
    ww->borc = colarray[COLBORD];
    ww->papr = 84;
    ww->curw = 1;
    ww->curc = 0;
    ww->uback = 7;
    ww->uink = 4;
    ww->aback = 7;
    ww->aink = 0;
    ww->sback = 4;
    ww->sink = 0;

    ww->ninfo = 2;
    ww->ninob = 1;
    ww->nappl = 1;

    ww->pappl = ww->splst = appl;
    ww->pinfo = infw;
    ww->plitm = ppl;
    ww->nlitm = 2;

    ws->wmode = FORM_QL4;
    ws->wwork = ww;		/* cross references */
    ww->wstat = ws;
    ww->chid = chid;

    ysize = m->height * 12;

    pr_w = strlen (m->title);
    ptitle = xalloc (pr_w + 8);
    ptitle->len = pr_w;
    memcpy (ptitle->chrs, m->title, pr_w);

    pr_w = (pr_w + 1) * 6;

    info[0].pobj = ptitle;
    info[0].xsize = pr_w - 6;	/* String length */

    /*-------------------------------------.
    | Don't exceed the screen size now...  |
    `-------------------------------------*/

    xsize = 4 + (m->width + EXTRAS) * 6;

    if (xsize < pr_w + 42)
    {
	xsize = pr_w + 42;
    }

    infw[0].xsize = xsize;
    infw[1].xsize = pr_w;	/* outer wight window */

    (ppl + 1)->xorg = xsize - (ppl + 1)->xsize - 4;

    /*-------------------------------------------------------.
    | load the PTR menu from the menu defined by the rc file |
    `-------------------------------------------------------*/

    LoadMenu (mw, m);

    ww->xsize = xsize;
    mw->xsize = xsize - 4;

    mw->ysize = ysize;
    ww->ysize = 15 + mw->ysize;
    mw->nysc = MAX_SECT;
    mw->ypwcb = (WM_pwcb_t *) & pwcy;
    ww->xorg = 4;
    ww->yorg = 15 + 8;

    ws->citem = -1;		/* no current item  */
    ws->ciact = NULL;		/* or current item routine */
    (appl + 2)->pappw = (WM_appw_t *) rbuf;

    /*----------------------------.
    | And draw the menu (at last) |
    `----------------------------*/
    signalTimer(0);
    wm_prpos (ww, -1, -1);
    wm_wdraw (ww);

    /*--------------------------------------------------------------------.
    | Now read the pointer, we will either get a CAN is the user pressed  |
    | ESC, or a DO if an invocable item was selected. This may have been  |
    | cascaded back from a sub menu, via the action routine.		  |
    `--------------------------------------------------------------------*/
    {

	while(1) 
	{
	    wm_rptr (ww);  
	    ualarm (0);
	    if (ws->evnt & PT_ZZZZ)
	    {
		menu_button(ww, "qfm");
		lit_reset(ww, 1);
	    }
	    else if (ws->evnt & PT_CAN)
	    {
		break;
	    }
	    else if (ws->evnt & PT_WMOVE)
	    {
		short dx, dy;
		struct QLRECT rct;
		wm_chwin (ww, &dx, &dy);   /* move window */
		lit_reset(ww, 0);
	    }
	    else if (ws->evnt & PT_DO)
	    {
		short i;
		ITEMS *im;		
		im = m->items;
		for (i = 0; i != *rbuf; i++, im = im->next)
		    ; /* */

		if(im && im->l->act)
		{
		    if(im->l->act == EXEC)
		    {
			char flag = (Mmstt && (*Mmstt == WSI_AVBL));
			if(execute(ww, rbuf, flag)) break;
			*(rbuf) = *(rbuf+1) = 0;
		    }
		    else if(im->l->act == XCSM)
		    {
			char args[40] = "";
			char tbuf[40] = "";
			int res;
			if(0 == GetQM(ww, 39, "QFAX CSM Command", args, tbuf))
			{
			    qfaxcsm(tbuf);
			}
		    }
		}
		else
		{
		    break;
		}
	    }
	    else if(Mmstt)
	    {
		wm_mdraw(ww, (WM_swdef_t *) mw, 0);
	    }
	    ualarm (50);
	}
    }
    /*-----------------------------------------------------------------.
    | The user has finished with the menu, we must clean up after them |
    `-----------------------------------------------------------------*/

    if (mw->rowl)
    {
	xfree (mw->rowl);
    }

    wm_cluns (ww);		/* throw it away */
    xfree (ppl);
    xfree (ptitle);
    xfree (info);
    xfree (infw);
    xfree (ww);
    xfree (mw);
    xfree (ws);
    return rval;
}

/*----------------------------.
| The draw callback for menus |
`----------------------------*/

static long MDRAW (WM_wwork_t * wwk, WM_menw_t * mw)
{

    menu_chks (mw);
    wm_index (wwk, (WM_swdef_t *) mw);
    wm_mdraw (wwk, (WM_swdef_t *) mw, 0);
}

/*-------------------------------------------------------.
| Reset an item in a menu after the user has selected it |
`-------------------------------------------------------*/

static void resetobj (WM_wwork_t * wwk, WM_menw_t * mw, char *mst)
{
    *mst = WSI_MKAV;
    wm_mdraw (wwk, (WM_swdef_t *) mw, -1);
    *mst = WSI_AVBL;
}

/*----------------------------------------------------------------------.
| This callback is called by the PE when the user clicks on a invocable |
| menu item.							        |
`----------------------------------------------------------------------*/

static long MEXEC (WM_wwork_t * wwk, WM_menw_t * mw, char *mst, short col,
		   short row)
{
    WM_mobj_t *mobj;
    void *q;
    MINFO *m;

    mobj = mw->rowl[row].rows;	/* current menu object */
    /*-------------------------------------------------------------.
    | This is where we've stashed away the pointer to the original |
    | callers data.						   |
    `-------------------------------------------------------------*/
    *(char*) ((wwk->pappl + 2)->pappw) = mobj->item;
    *((char*) ((wwk->pappl + 2)->pappw) + 1) = 	
	(wwk->wstat->kprs == K_DO) ? 'r' : 'l';
    resetobj (wwk, mw, mst + mobj->item);
    return PT__DO;
}
